@@ -8,7 +8,6 @@
 {-# LANGUAGE QuasiQuotes #-}
 {-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE TypeFamilies #-}
-
 -- redundant-constraints: False positive in default signature for `toRawValue`
 {-# OPTIONS_GHC -Wno-redundant-constraints #-}
 
@@ -34,7 +33,6 @@ module Hercules.CNix.Expr
     evalFile,
     newStrings,
     appendString,
-    evalArgs,
     autoCallFunction,
     isDerivation,
     isFunctor,
@@ -54,10 +52,10 @@ module Hercules.CNix.Expr
     getLocalFlake,
     getFlakeFromGit,
     getFlakeFromArchiveUrl,
-    ToRawValue(..),
-    ToValue(..),
-    FromValue(..),
-    ViaJSON(..),
+    ToRawValue (..),
+    ToValue (..),
+    FromValue (..),
+    ViaJSON (..),
 
     -- * Re-exports
     RawValue,
@@ -74,6 +72,8 @@ where
 
 import Conduit
 import qualified Data.Aeson as A
+import Data.Aeson.KeyMap (toMapText)
+import qualified Data.ByteString.Unsafe as BS
 import Data.Coerce (coerce)
 import qualified Data.HashMap.Lazy as H
 import qualified Data.Map as M
@@ -82,6 +82,7 @@ import Data.Vector (Vector)
 import qualified Data.Vector as V
 import Foreign (nullPtr)
 import qualified Foreign.C.String
+import qualified Hercules.CNix as CNix
 import Hercules.CNix.Encapsulation (moveToForeignPtrWrapper)
 import Hercules.CNix.Expr.Context
 import Hercules.CNix.Expr.Raw
@@ -90,10 +91,8 @@ import Hercules.CNix.Store
 import Hercules.CNix.Store.Context
 import qualified Language.C.Inline.Cpp as C
 import qualified Language.C.Inline.Cpp.Exception as C
-import Paths_hercules_ci_cnix_expr (getDataFileName)
 import Protolude hiding (evalState)
 import System.Directory (makeAbsolute)
-import Data.Aeson.KeyMap (toMapText)
 
 C.context (Hercules.CNix.Store.Context.context <> Hercules.CNix.Expr.Context.evalContext)
 
@@ -105,34 +104,22 @@ C.include "<cstring>"
 
 C.include "<math.h>"
 
-C.include "<nix/config.h>"
-
-C.include "<nix/shared.hh>"
-
-C.include "<nix/eval.hh>"
-
-C.include "<nix/eval-inline.hh>"
-
-C.include "<nix/store-api.hh>"
-
-C.include "<nix/common-eval-args.hh>"
-
-C.include "<nix/get-drvs.hh>"
-
-C.include "<nix/derivations.hh>"
-
-C.include "<nix/globals.hh>"
-
+C.include "<nix/util/config-global.hh>"
+C.include "<nix/store/store-api.hh>"
+C.include "<nix/store/globals.hh>"
+C.include "<nix/expr/eval.hh>"
+C.include "<nix/expr/eval-inline.hh>"
+C.include "<nix/expr/get-drvs.hh>"
 C.include "<nix/flake/flake.hh>"
-
 C.include "<nix/flake/flakeref.hh>"
-
-#if NIX_IS_AT_LEAST(2,19,0)
-C.include "<nix/args/root.hh>"
-#endif
+C.include "<nix/flake/settings.hh>"
+C.include "<nix/fetchers/fetch-settings.hh>"
+C.include "<nix/main/shared.hh>"
 
 C.include "hercules-ci-cnix/expr.hxx"
 
+C.include "hercules-ci-cnix/string.hxx"
+
 C.include "<gc/gc.h>"
 
 C.include "<gc/gc_cpp.h>"
@@ -141,27 +128,28 @@ C.include "<gc/gc_allocator.h>"
 
 C.using "namespace nix"
 
+C.using "namespace hercules_ci_cnix"
+
 C.verbatim "\nGC_API void GC_CALL GC_throw_bad_alloc() { throw std::bad_alloc(); }\n"
 
+C.verbatim "nix::flake::Settings flakeSettings;"
+C.verbatim "nix::fetchers::Settings fetchSettings;"
+C.verbatim "static bool readOnlyMode = false;"
+C.verbatim "nix::EvalSettings evalSettings(readOnlyMode);"
+
 init :: IO ()
-init =
-  void
-    [C.throwBlock| void {
-      nix::initNix();
-      nix::initGC();
-#if NIX_IS_AT_LEAST(2,15,0)
-      globalConfig.set("extra-experimental-features", "flakes");
-#else
-#if NIX_IS_AT_LEAST(2,5,0)
-      std::set<nix::ExperimentalFeature> features(nix::settings.experimentalFeatures.get());
-      features.insert(nix::ExperimentalFeature::Flakes);
-#else
-      Strings features(nix::settings.experimentalFeatures.get());
-      features.push_back("flakes");
-#endif
-      nix::settings.experimentalFeatures.assign(features);
-#endif
-    } |]
+init = do
+  [C.throwBlock| void {
+    { auto _ = GlobalConfig::Register(&flakeSettings); }
+    { auto _ = GlobalConfig::Register(&fetchSettings); }
+    { auto _ = GlobalConfig::Register(&evalSettings); }
+  }|]
+  CNix.init
+  [C.throwBlock| void {
+    nix::initGC();
+    globalConfig.set("extra-experimental-features", "flakes");
+    flakeSettings.configureEvalSettings(evalSettings);
+  }|]
 
 -- | Initialize the current (main) thread for stack overflow detection.
 initThread :: IO ()
@@ -211,15 +199,12 @@ runGcRegisteredThread io =
         }|]
         pass
 
-
-{- | Configure the stack overflow handler to sleep before returning, allowing
-  other threads to continue for a bit.
-
-  No-op before Nix 2.12
--}
+-- | Configure the stack overflow handler to sleep before returning, allowing
+--  other threads to continue for a bit.
+--
+--  No-op before Nix 2.12
 setExtraStackOverflowHandlerToSleep :: IO ()
 setExtraStackOverflowHandlerToSleep =
-#if NIX_IS_AT_LEAST(2,12,0)
   void
     [C.throwBlock| void {
       nix::stackOverflowHandler = [](siginfo_t *info, void *ctx) {
@@ -234,9 +219,6 @@ setExtraStackOverflowHandlerToSleep =
         _exit(1);
         };
     }|]
-#else
-  pass
-#endif
 
 setTalkative :: IO ()
 setTalkative =
@@ -274,22 +256,16 @@ logInfo t = do
   }|]
 
 -- | (private) Make an EvalState and leak it.
-newEvalState :: MonadIO m => Store -> m (Ptr EvalState)
-newEvalState (Store store) = liftIO
-#if NIX_IS_AT_LEAST(2,17,0)
-  [C.throwBlock| EvalState* {
-    nix::SearchPath searchPaths;
-    return new EvalState(searchPaths, *$(refStore* store));
+newEvalState :: (MonadIO m) => Store -> m (Ptr EvalState)
+newEvalState (Store store) =
+  liftIO
+    [C.throwBlock| EvalState* {
+      nix::LookupPath emptyLookupPath;
+      return new EvalState(emptyLookupPath, *$(refStore* store), fetchSettings, evalSettings);
   } |]
-#else
-  [C.throwBlock| EvalState* {
-    Strings searchPaths;
-    return new EvalState(searchPaths, *$(refStore* store));
-  } |]
-#endif
 
 -- | (private) Don't leak it.
-deleteEvalState :: MonadIO m => (Ptr EvalState) -> m ()
+deleteEvalState :: (MonadIO m) => (Ptr EvalState) -> m ()
 deleteEvalState st = liftIO [C.throwBlock| void { delete $(EvalState* st); } |]
 
 withEvalState ::
@@ -299,30 +275,30 @@ withEvalState ::
 withEvalState store = bracket (newEvalState store) deleteEvalState
 
 withEvalStateConduit ::
-  MonadResource m =>
+  (MonadResource m) =>
   Store ->
   (Ptr EvalState -> ConduitT i o m r) ->
   ConduitT i o m r
 withEvalStateConduit store = bracketP (newEvalState store) deleteEvalState
 
 -- | Insert an allowed path. Only has an effect when in restricted or pure mode.
+{- ORMOLU_DISABLE -}
 addAllowedPath :: Ptr EvalState -> ByteString -> IO ()
 addAllowedPath evalState path =
   [C.throwBlock| void {
     std::string path = std::string($bs-ptr:path, $bs-len:path);
     EvalState &evalState = *$(EvalState *evalState);
-#if NIX_IS_AT_LEAST(2,20,0)
-    evalState.allowPath(path);
+#if NIX_IS_AT_LEAST(2, 32, 0)
+    evalState.allowPathLegacy(path);
 #else
-    if (evalState.allowedPaths) {
-      evalState.allowedPaths->insert(path);
-    }
+    evalState.allowPath(path);
 #endif
   }|]
+{- ORMOLU_ENABLE -}
 
 addInternalAllowedPaths :: Ptr EvalState -> IO ()
-addInternalAllowedPaths evalState = do
-  addAllowedPath evalState . encodeUtf8 . toS =<< getDataFileName "vendor/flake-compat"
+addInternalAllowedPaths _evalState = do
+  pass
 
 evalFile :: Ptr EvalState -> FilePath -> IO RawValue
 evalFile evalState filename = do
@@ -332,13 +308,7 @@ evalFile evalState filename = do
       EvalState & state = *$(EvalState *evalState);
       Value value;
       auto cstr = $(const char *filename');
-#if NIX_IS_AT_LEAST(2,19,0)
       SourcePath path {state.rootPath(CanonPath(cstr))};
-#elif NIX_IS_AT_LEAST(2,16,0)
-      SourcePath path = CanonPath(cstr);
-#else
-      std::string path = cstr;
-#endif
       state.evalFile(path, value);
       return new (NoGC) Value(value);
     }|]
@@ -353,40 +323,13 @@ appendString ss s =
     $(Strings *ss)->push_back(std::string($bs-ptr:s, $bs-len:s));
   }|]
 
-#if NIX_IS_AT_LEAST(2,19,0)
-C.verbatim "struct EvalArgs : nix::RootArgs, nix::MixEvalArgs { };"
-#else
-C.verbatim "struct EvalArgs : nix::MixEvalArgs { };"
-#endif
-
-evalArgs :: Ptr EvalState -> [ByteString] -> IO (Value NixAttrs)
-evalArgs evalState args = do
-  argsStrings <- newStrings
-  forM_ args $ appendString argsStrings
-  fmap unsafeAssertType . mkRawValue
-    =<< [C.throwBlock| Value * {
-      Strings *args = $(Strings *argsStrings);
-      struct EvalArgs evalArgs;
-      Bindings *autoArgs;
-      EvalState &state = *$(EvalState *evalState);
-
-      evalArgs.parseCmdline(*args);
-      autoArgs = evalArgs.getAutoArgs(state);
-      if (!autoArgs) {
-        throw nix::Error("Could not evaluate automatic arguments");
-      }
-      Value *r = new (NoGC) Value ();
-      r->mkAttrs(autoArgs);
-      return r;
-    }|]
-
 autoCallFunction :: Ptr EvalState -> RawValue -> Value NixAttrs -> IO RawValue
 autoCallFunction evalState (RawValue fun) (Value (RawValue autoArgs)) =
   mkRawValue
     =<< [C.throwBlock| Value* {
           Value result;
           $(EvalState *evalState)->autoCallFunction(
-                  *$(Value *autoArgs)->attrs,
+                  const_cast<Bindings &>(*$(Value *autoArgs)->attrs()),
                   *$(Value *fun),
                   result);
           return new (NoGC) Value (result);
@@ -409,30 +352,38 @@ isFunctor evalState (RawValue v) =
           return $(EvalState *evalState)->isFunctor(*$(Value *v));
         }|]
 
+{- ORMOLU_DISABLE -}
 getRecurseForDerivations :: Ptr EvalState -> Value NixAttrs -> IO Bool
 getRecurseForDerivations evalState (Value (RawValue v)) =
   (0 /=)
     <$> [C.throwBlock| int {
           Value *v = $(Value *v);
           EvalState &evalState = *$(EvalState *evalState);
-          Bindings::iterator iter = v->attrs->find(evalState.sRecurseForDerivations);
-          if (iter == v->attrs->end()) {
+          auto attrs = v->attrs();
+#if NIX_IS_AT_LEAST(2, 32, 0)
+          auto iter = attrs->get(evalState.s.recurseForDerivations);
+          if (!iter) {
+            return 0;
+          } else {
+            return evalState.forceBool(*iter->value, iter->pos, "while evaluating whether to traverse into an attribute set to find more derivations");
+          }
+#else
+          // Bindings::const_iterator iter; // 2.24
+          auto iter = attrs->find(evalState.sRecurseForDerivations);
+          if (iter == attrs->end()) {
             return 0;
           } else {
             // Previously this bool was unpacked manually and included a special
             // case to return true when it is not a bool. That logic was added
             // because an empty attrset was found here, observed in
             // nixpkgs master 67e2de195a4aa0a50ffb1e1ba0b4fb531dca67dc
-#if NIX_IS_AT_LEAST(2,14,0)
             return evalState.forceBool(*iter->value, iter->pos, "while evaluating whether to traverse into an attribute set to find more derivations");
-#elif NIX_IS_AT_LEAST(2,9,0)
-            return evalState.forceBool(*iter->value, iter->pos);
-#else
-            return evalState.forceBool(*iter->value, *iter->pos);
-#endif
           }
+#endif
         } |]
+{- ORMOLU_ENABLE -}
 
+{- ORMOLU_DISABLE -}
 getAttr :: Ptr EvalState -> Value NixAttrs -> ByteString -> IO (Maybe RawValue)
 getAttr evalState (Value (RawValue v)) k =
   mkNullableRawValue
@@ -440,56 +391,89 @@ getAttr evalState (Value (RawValue v)) k =
       Value &v = *$(Value *v);
       EvalState &evalState = *$(EvalState *evalState);
       Symbol k = evalState.symbols.create($bs-cstr:k);
-      Bindings::iterator iter = v.attrs->find(k);
-      if (iter == v.attrs->end()) {
+      auto attrs = v.attrs();
+#if NIX_IS_AT_LEAST(2, 32, 0)
+      auto iter = attrs->get(k);
+      if (!iter) {
         return nullptr;
       } else {
         return iter->value;
       }
+#else
+      // Bindings::const_iterator iter; // 2.24
+      auto iter = attrs->find(k);
+      if (iter == attrs->end()) {
+        return nullptr;
+      } else {
+        return iter->value;
+      }
+#endif
     }|]
+{- ORMOLU_ENABLE -}
 
 -- | Converts 'nullPtr' to 'Nothing'; actual values to @Just (a :: 'RawValue')@
 mkNullableRawValue :: Ptr Value' -> IO (Maybe RawValue)
 mkNullableRawValue p | p == nullPtr = pure Nothing
 mkNullableRawValue p = Just <$> mkRawValue p
 
+{- ORMOLU_DISABLE -}
 getAttrs :: Ptr EvalState -> Value NixAttrs -> IO (Map ByteString RawValue)
 getAttrs evalState (Value (RawValue v)) = do
-  begin <- [C.exp| Attr *{ $(Value *v)->attrs->begin() }|]
-  end <- [C.exp| Attr *{ $(Value *v)->attrs->end() }|]
+#if NIX_IS_AT_LEAST(2, 32, 0)
+  -- In Nix 2.32+, we need to collect all attrs in C++ since iteration changed
+  size <- [C.exp| size_t { $(Value *v)->attrs()->size() }|]
+  let collect :: C.CSize -> Map ByteString RawValue -> IO (Map ByteString RawValue)
+      collect i acc | i >= size = pure acc
+      collect i acc = do
+        name <-
+          BS.unsafePackMallocCString
+            =<< [C.block| const char *{
+              EvalState &evalState = *$(EvalState *evalState);
+              auto iter = $(Value *v)->attrs()->begin();
+              for (size_t j = 0; j < $(size_t i); j++) ++iter;
+              SymbolStr str = evalState.symbols[iter->name];
+              return stringdup(static_cast<std::string>(str));
+            }|]
+        value <- mkRawValue =<< [C.block| Value *{
+              auto iter = $(Value *v)->attrs()->begin();
+              for (size_t j = 0; j < $(size_t i); j++) ++iter;
+              return new (NoGC) Value(*iter->value);
+            }|]
+        let acc' = M.insert name value acc
+        seq acc' pass
+        collect (i + 1) acc'
+  collect 0 mempty
+#else
+  begin <- [C.exp| const Attr *{ $(Value *v)->attrs()->begin() }|]
+  end <- [C.exp| const Attr *{ $(Value *v)->attrs()->end() }|]
   let gather :: Map ByteString RawValue -> Ptr Attr' -> IO (Map ByteString RawValue)
       gather acc i | i == end = pure acc
       gather acc i = do
-#if NIX_IS_AT_LEAST(2,9,0)
-        name <- unsafeMallocBS [C.block| const char *{
-          EvalState &evalState = *$(EvalState *evalState);
-          SymbolStr str = evalState.symbols[$(Attr *i)->name];
-          return strdup(static_cast<std::string>(str).c_str());
-        }|]
-#else
-        name <- unsafeMallocBS [C.exp| const char *{ strdup(static_cast<std::string>($(Attr *i)->name).c_str()) } |]
-#endif
+        name <-
+          BS.unsafePackMallocCString
+            =<< [C.block| const char *{
+              EvalState &evalState = *$(EvalState *evalState);
+              SymbolStr str = evalState.symbols[$(Attr *i)->name];
+              return stringdup(static_cast<std::string>(str));
+            }|]
         value <- mkRawValue =<< [C.exp| Value *{ new (NoGC) Value(*$(Attr *i)->value) } |]
         let acc' = M.insert name value acc
         seq acc' pass
         gather acc' =<< [C.exp| Attr *{ &$(Attr *i)[1] }|]
   gather mempty begin
+#endif
+{- ORMOLU_ENABLE -}
 
-getDrvFile :: MonadIO m => Ptr EvalState -> RawValue -> m StorePath
+getDrvFile :: (MonadIO m) => Ptr EvalState -> RawValue -> m StorePath
 getDrvFile evalState (RawValue v) = liftIO do
   moveToForeignPtrWrapper
     =<< [C.throwBlock| nix::StorePath *{
       EvalState &state = *$(EvalState *evalState);
       auto drvInfo = getDerivation(state, *$(Value *v), false);
       if (!drvInfo)
-        throw EvalError("Not a valid derivation");
+        throw EvalError(state, "Not a valid derivation");
 
-#if NIX_IS_AT_LEAST(2,7,0)
       StorePath storePath = drvInfo->requireDrvPath();
-#else
-      std::string drvPath = drvInfo->queryDrvPath();
-      StorePath storePath = state.store->parseStorePath(drvPath);
-#endif
 
       // write it (?)
       auto drv = state.store->derivationFromPath(storePath);
@@ -510,12 +494,26 @@ getAttrBool evalState attrset attrName = do
       Right _ -> do
         pure $ Right Nothing
 
+#if NIX_IS_AT_LEAST(2, 30, 0)
+getList :: Value NixList -> IO [RawValue]
+getList (Value (RawValue nixList)) = do
+  len <- [C.exp| int { $(Value *nixList)->listSize() }|]
+  bracket
+    [C.exp| ListViewType * { new ListViewType($(Value *nixList)->listView()) }|]
+    (\listView -> [C.exp| void { delete $(ListViewType *listView) }|])
+    (\listView -> do
+      let getElem i = mkRawValue =<< [C.exp| Value * { (*$(ListViewType *listView))[$(int i)] }|]
+      for [0 .. (len - 1)] \i -> do
+        getElem i
+    )
+#else
 getList :: Value NixList -> IO [RawValue]
 getList (Value (RawValue nixList)) = do
   len <- [C.exp| int { $(Value *nixList)->listSize() }|]
   let getElem i = mkRawValue =<< [C.exp| Value * { $(Value *nixList)->listElems()[$(int i)] }|]
   for [0 .. (len - 1)] \i -> do
     getElem i
+#endif
 
 getAttrList :: Ptr EvalState -> Value NixAttrs -> ByteString -> IO (Either SomeException (Maybe [RawValue]))
 getAttrList evalState attrset attrName = do
@@ -543,13 +541,7 @@ valueFromExpressionString evalState s basePath = do
     =<< [C.throwBlock| Value *{
       EvalState &evalState = *$(EvalState *evalState);
       std::string basePathStr = std::string($bs-ptr:basePath, $bs-len:basePath);
-#if NIX_IS_AT_LEAST(2,19,0)
       SourcePath basePath {evalState.rootPath(CanonPath(basePathStr))};
-#elif NIX_IS_AT_LEAST(2,16,0)
-      SourcePath basePath = CanonPath(basePathStr);
-#else
-      auto & basePath = basePathStr;
-#endif
       Expr *expr = evalState.parseExprFromString(std::string($bs-ptr:s, $bs-len:s), basePath);
       Value *r = new (NoGC) Value();
       evalState.eval(expr, *r);
@@ -576,22 +568,24 @@ apply (RawValue f) (RawValue a) = do
       return r;
     }|]
 
+{- ORMOLU_DISABLE -}
 mkPath :: Ptr EvalState -> ByteString -> IO (Value NixPath)
 mkPath evalState path =
   Value
     <$> ( mkRawValue
             =<< [C.throwBlock| Value *{
-      EvalState & state = *$(EvalState *evalState);
-      Value *r = new (NoGC) Value();
-      std::string s($bs-ptr:path, $bs-len:path);
-#if NIX_IS_AT_LEAST(2,19,0)
-      r->mkPath(state.rootPath(CanonPath(s)));
+              Value *r = new (NoGC) Value();
+              std::string s($bs-ptr:path, $bs-len:path);
+              EvalState & state = *$(EvalState *evalState);
+#if NIX_IS_AT_LEAST(2, 33, 0)
+              r->mkPath(state.rootPath(CanonPath(s)), state.mem);
 #else
-      r->mkPath(s.c_str());
+              r->mkPath(state.rootPath(CanonPath(s)));
 #endif
-      return r;
-  }|]
+              return r;
+          }|]
         )
+{- ORMOLU_ENABLE -}
 
 getFlakeFromFlakeRef :: Ptr EvalState -> ByteString -> IO RawValue
 getFlakeFromFlakeRef evalState flakeRef = do
@@ -599,17 +593,20 @@ getFlakeFromFlakeRef evalState flakeRef = do
     EvalState &evalState = *$(EvalState *evalState);
     Value *r = new (NoGC) Value();
     std::string flakeRefStr($bs-ptr:flakeRef, $bs-len:flakeRef);
-    auto flakeRef = nix::parseFlakeRef(flakeRefStr, {}, true);
+    auto flakeRef = nix::parseFlakeRef(
+      fetchSettings,
+      flakeRefStr,
+      {},
+      true);
     nix::flake::callFlake(evalState,
-      nix::flake::lockFlake(evalState, flakeRef,
+      nix::flake::lockFlake(
+        flakeSettings,
+        evalState,
+        flakeRef,
         nix::flake::LockFlags {
           .updateLockFile = false,
           .useRegistries = false,
-#if NIX_IS_AT_LEAST(2,13,0)
           .allowUnlocked = false,
-#else
-          .allowMutable = false,
-#endif
         }),
       *r);
     return r;
@@ -621,32 +618,34 @@ getLocalFlake evalState path = do
   absPath <- encodeUtf8 . toS <$> makeAbsolute (toS path)
   mkRawValue
     =<< [C.throwBlock| Value *{
-    EvalState &evalState = *$(EvalState *evalState);
-    Value *r = new (NoGC) Value();
-    std::string path($bs-ptr:absPath, $bs-len:absPath);
-    auto flakeRef = nix::parseFlakeRef(path, {}, true);
-    nix::flake::callFlake(evalState,
-      nix::flake::lockFlake(evalState, flakeRef,
-        nix::flake::LockFlags {
-          .updateLockFile = false,
-          .useRegistries = false,
-#if NIX_IS_AT_LEAST(2,13,0)
-          .allowUnlocked = false,
-#else
-          .allowMutable = false,
-#endif
-        }),
-      *r);
-    return r;
-  }|]
+      EvalState &evalState = *$(EvalState *evalState);
+      Value *r = new (NoGC) Value();
+      std::string path($bs-ptr:absPath, $bs-len:absPath);
+      auto flakeRef = nix::parseFlakeRef(
+        fetchSettings,
+        path,
+        {},
+        true);
+      nix::flake::callFlake(evalState,
+        nix::flake::lockFlake(
+          flakeSettings,
+          evalState,
+          flakeRef,
+          nix::flake::LockFlags {
+            .updateLockFile = false,
+            .useRegistries = false,
+            .allowUnlocked = false,
+          }),
+        *r);
+      return r;
+    }|]
 
 getFlakeFromGit :: Ptr EvalState -> Text -> Text -> Text -> IO RawValue
 getFlakeFromGit evalState url ref rev =
-  let
-    urlb = encodeUtf8 url
-    refb = encodeUtf8 ref
-    revb = encodeUtf8 rev
-  in [C.throwBlock| Value *{
+  let urlb = encodeUtf8 url
+      refb = encodeUtf8 ref
+      revb = encodeUtf8 rev
+   in [C.throwBlock| Value *{
     EvalState &evalState = *$(EvalState *evalState);
     Value *r = new (NoGC) Value();
     std::string url($bs-ptr:urlb, $bs-len:urlb);
@@ -659,22 +658,23 @@ getFlakeFromGit evalState url ref rev =
     attrs.emplace("ref", ref);
     attrs.emplace("rev", rev);
 
-    auto flakeRef = nix::FlakeRef::fromAttrs(attrs);
+    auto flakeRef = nix::FlakeRef::fromAttrs(
+      fetchSettings,
+      attrs);
     nix::flake::callFlake(evalState,
-      nix::flake::lockFlake(evalState, flakeRef,
+      nix::flake::lockFlake(
+        flakeSettings,
+        evalState,
+        flakeRef,
         nix::flake::LockFlags {
           .updateLockFile = false,
           .useRegistries = false,
-#if NIX_IS_AT_LEAST(2,13,0)
           .allowUnlocked = false,
-#else
-          .allowMutable = false,
-#endif
         }),
       *r);
     return r;
   }|]
-    >>= mkRawValue
+        >>= mkRawValue
 
 getFlakeFromArchiveUrl :: Ptr EvalState -> Text -> IO RawValue
 getFlakeFromArchiveUrl evalState url = do
@@ -687,15 +687,15 @@ getFlakeFromArchiveUrl evalState url = do
   p' <- getStringIgnoreContext p
   getFlakeFromFlakeRef evalState p'
 
-traverseWithKey_ :: Applicative f => (k -> a -> f ()) -> Map k a -> f ()
+traverseWithKey_ :: (Applicative f) => (k -> a -> f ()) -> Map k a -> f ()
 traverseWithKey_ f = M.foldrWithKey (\k a more -> f k a *> more) (pure ())
 
 class ToRawValue a where
   toRawValue :: Ptr EvalState -> a -> IO RawValue
-  default toRawValue :: ToValue a => Ptr EvalState -> a -> IO RawValue
+  default toRawValue :: (ToValue a) => Ptr EvalState -> a -> IO RawValue
   toRawValue evalState a = rtValue <$> toValue evalState a
 
-class ToRawValue a => ToValue a where
+class (ToRawValue a) => ToValue a where
   type NixTypeFor a :: Type
   toValue :: Ptr EvalState -> a -> IO (Value (NixTypeFor a))
 
@@ -735,10 +735,10 @@ instance ToValue C.CBool where
   toValue _ b =
     coerce
       <$> [C.block| Value *{
-      Value *r = new (NoGC) Value();
-      r->mkBool($(bool b));
-      return r;
-    }|]
+        Value *r = new (NoGC) Value();
+        r->mkBool($(bool b));
+        return r;
+      }|]
 
 instance ToRawValue Bool
 
@@ -756,10 +756,10 @@ instance ToValue Int64 where
   toValue _ i =
     coerce
       <$> [C.block| Value *{
-    Value *r = new (NoGC) Value();
-    r->mkInt($(int64_t i));
-    return r;
-  }|]
+        Value *r = new (NoGC) Value();
+        r->mkInt($(int64_t i));
+        return r;
+      }|]
 
 instance ToRawValue Int
 
@@ -774,10 +774,10 @@ instance ToValue C.CDouble where
   toValue _ f =
     coerce
       <$> [C.block| Value *{
-        Value *r = new (NoGC) Value();
-        r->mkFloat($(double f));
-        return r;
-      }|]
+          Value *r = new (NoGC) Value();
+          r->mkFloat($(double f));
+          return r;
+        }|]
 
 instance ToRawValue Double
 
@@ -785,24 +785,36 @@ instance ToValue Double where
   type NixTypeFor Double = NixFloat
   toValue es f = toValue es (fromRational (toRational f) :: C.CDouble)
 
+{- ORMOLU_DISABLE -}
 -- | Nix String
 instance ToValue ByteString where
   type NixTypeFor ByteString = NixString
-  toValue _ s =
+  toValue evalState s =
     -- TODO simplify when r->mkString(string_view) is safe in all supported Nix versions
     coerce
       <$> [C.block| Value *{
-    Value *r = new (NoGC) Value();
-    std::string_view s($bs-ptr:s, $bs-len:s);
-    // If empty, the pointer may be invalid; don't use it.
-    if (s.size() == 0) {
-      r->mkString("");
-    }
-    else {
-      r->mkString(GC_STRNDUP(s.data(), s.size()));
-    }
-    return r;
-  }|]
+        Value *r = new (NoGC) Value();
+        std::string_view s($bs-ptr:s, $bs-len:s);
+#if NIX_IS_AT_LEAST(2, 33, 0)
+        EvalState &evalState = *$(EvalState *evalState);
+        if (s.size() == 0) {
+          r->mkString(std::string_view(""), evalState.mem);
+        }
+        else {
+          r->mkString(std::string_view(GC_STRNDUP(s.data(), s.size()), s.size()), evalState.mem);
+        }
+#else
+        // If empty, the pointer may be invalid; don't use it.
+        if (s.size() == 0) {
+          r->mkString("");
+        }
+        else {
+          r->mkString(GC_STRNDUP(s.data(), s.size()));
+        }
+#endif
+        return r;
+      }|]
+{- ORMOLU_ENABLE -}
 
 -- | Nix String
 instance ToRawValue ByteString
@@ -815,74 +827,53 @@ instance ToValue Text where
   type NixTypeFor Text = NixString
   toValue es s = toValue es (encodeUtf8 s)
 
-instance ToRawValue a => ToRawValue (Map ByteString a)
+instance (ToRawValue a) => ToRawValue (Map ByteString a)
 
-#if NIX_IS_AT_LEAST(2,6,0)
-withBindingsBuilder :: Integral n => Ptr EvalState -> n -> (Ptr BindingsBuilder' -> IO ()) -> IO (Value NixAttrs)
+withBindingsBuilder :: (Integral n) => Ptr EvalState -> n -> (Ptr BindingsBuilder' -> IO ()) -> IO (Value NixAttrs)
 withBindingsBuilder evalState n f = do
   withBindingsBuilder' evalState n \bb -> do
     f bb
-    v <- [C.block| Value* {
-      auto v = new (NoGC) Value();
-      v->mkAttrs(*$(BindingsBuilder *bb));
-      return v;
-    }|]
+    v <-
+      [C.block| Value* {
+        auto v = new (NoGC) Value();
+        v->mkAttrs(*$(BindingsBuilder *bb));
+        return v;
+      }|]
     Value <$> mkRawValue v
 
-withBindingsBuilder' :: Integral n => Ptr EvalState -> n -> (Ptr BindingsBuilder' -> IO a) -> IO a
+{- ORMOLU_DISABLE -}
+withBindingsBuilder' :: (Integral n) => Ptr EvalState -> n -> (Ptr BindingsBuilder' -> IO a) -> IO a
 withBindingsBuilder' evalState n =
   let l :: C.CInt
       l = fromIntegral n
-  in
-    bracket
-      [C.block| BindingsBuilder* {
-        auto &evalState = *$(EvalState *evalState);
-        return new BindingsBuilder(evalState, evalState.allocBindings($(int l)));
-      }|]
-      \bb -> [C.block| void { delete $(BindingsBuilder *bb); }|]
+   in bracket
+        [C.block| BindingsBuilder* {
+          auto &evalState = *$(EvalState *evalState);
+#if NIX_IS_AT_LEAST(2, 32, 0)
+          return new BindingsBuilder(evalState.buildBindings($(int l)));
+#else
+          return new BindingsBuilder(evalState, evalState.allocBindings($(int l)));
 #endif
+        }|]
+        \bb -> [C.block| void { delete $(BindingsBuilder *bb); }|]
+{- ORMOLU_ENABLE -}
 
-instance ToRawValue a => ToValue (Map ByteString a) where
+instance (ToRawValue a) => ToValue (Map ByteString a) where
   type NixTypeFor (Map ByteString a) = NixAttrs
 
-#if NIX_IS_AT_LEAST(2,6,0)
   toValue evalState attrs = withBindingsBuilder evalState (length attrs) \bb -> do
     attrs & traverseWithKey_ \k a -> do
       RawValue aRaw <- toRawValue evalState a
       [C.block| void {
-          EvalState &evalState = *$(EvalState *evalState);
-          std::string k($bs-ptr:k, $bs-len:k);
-          Value &a = *$(Value *aRaw);
-          $(BindingsBuilder *bb)->alloc(evalState.symbols.create(k)) = a;
-        }|]
-#else
-  toValue evalState attrs = do
-    let l :: C.CInt
-        l = fromIntegral (length attrs)
-    v <-
-      [C.block| Value* {
-          EvalState &evalState = *$(EvalState *evalState);
-          Value *v = new (NoGC) Value();
-          evalState.mkAttrs(*v, $(int l));
-          return v;
-        }|]
-    attrs & traverseWithKey_ \k a -> do
-      RawValue aRaw <- toRawValue evalState a
-      [C.block| void {
-          EvalState &evalState = *$(EvalState *evalState);
-          std::string k($bs-ptr:k, $bs-len:k);
-          Value &a = *$(Value *aRaw);
-          *evalState.allocAttr(*$(Value *v), evalState.symbols.create(k)) = a;
-        }|]
-    [C.block| void {
-        $(Value *v)->attrs->sort();
+        EvalState &evalState = *$(EvalState *evalState);
+        std::string k($bs-ptr:k, $bs-len:k);
+        Value &a = *$(Value *aRaw);
+        $(BindingsBuilder *bb)->alloc(evalState.symbols.create(k)) = a;
       }|]
-    Value <$> mkRawValue v
-#endif
 
-instance ToRawValue a => ToRawValue (Map Text a)
+instance (ToRawValue a) => ToRawValue (Map Text a)
 
-instance ToRawValue a => ToValue (Map Text a) where
+instance (ToRawValue a) => ToValue (Map Text a) where
   type NixTypeFor (Map Text a) = NixAttrs
   toValue evalState attrs = toValue evalState (M.mapKeys encodeUtf8 attrs)
 
@@ -908,80 +899,67 @@ instance ToRawValue A.Value where
 newtype ViaJSON a = ViaJSON {fromViaJSON :: a}
   deriving newtype (Eq, Ord, Read, Show)
 
-instance A.ToJSON a => ToRawValue (ViaJSON a) where
+instance (A.ToJSON a) => ToRawValue (ViaJSON a) where
   toRawValue es (ViaJSON a) = toRawValue es (A.toJSON a)
 
-hmTraverseWithKey_ :: Applicative f => (k -> a -> f ()) -> H.HashMap k a -> f ()
+hmTraverseWithKey_ :: (Applicative f) => (k -> a -> f ()) -> H.HashMap k a -> f ()
 hmTraverseWithKey_ f = H.foldrWithKey (\k a more -> f k a *> more) (pure ())
 
-instance ToRawValue a => ToRawValue (H.HashMap Text a)
+instance (ToRawValue a) => ToRawValue (H.HashMap Text a)
 
-instance ToRawValue a => ToValue (H.HashMap Text a) where
+instance (ToRawValue a) => ToValue (H.HashMap Text a) where
   type NixTypeFor (H.HashMap Text a) = NixAttrs
 
-#if NIX_IS_AT_LEAST(2,6,0)
   toValue evalState attrs = withBindingsBuilder evalState (length attrs) \bb -> do
     attrs & hmTraverseWithKey_ \k' a -> do
       RawValue aRaw <- toRawValue evalState a
       let k = encodeUtf8 k'
       [C.block| void {
-          EvalState &evalState = *$(EvalState *evalState);
-          std::string k($bs-ptr:k, $bs-len:k);
-          Value &a = *$(Value *aRaw);
-          $(BindingsBuilder *bb)->alloc(evalState.symbols.create(k)) = a;
-        }|]
-#else
-  toValue evalState attrs = do
-    let l :: C.CInt
-        l = fromIntegral (length attrs)
-    v <-
-      [C.block| Value* {
-          EvalState &evalState = *$(EvalState *evalState);
-          Value *v = new (NoGC) Value();
-          evalState.mkAttrs(*v, $(int l));
-          return v;
-        }|]
-    attrs & hmTraverseWithKey_ \k' a -> do
-      RawValue aRaw <- toRawValue evalState a
-      let k = encodeUtf8 k'
-      [C.block| void {
-          EvalState &evalState = *$(EvalState *evalState);
-          std::string k($bs-ptr:k, $bs-len:k);
-          Value &a = *$(Value *aRaw);
-          *evalState.allocAttr(*$(Value *v), evalState.symbols.create(k)) = a;
-        }|]
-    [C.block| void {
-        $(Value *v)->attrs->sort();
+        EvalState &evalState = *$(EvalState *evalState);
+        std::string k($bs-ptr:k, $bs-len:k);
+        Value &a = *$(Value *aRaw);
+        $(BindingsBuilder *bb)->alloc(evalState.symbols.create(k)) = a;
       }|]
-    Value <$> mkRawValue v
-#endif
 
-instance ToRawValue a => ToRawValue (Vector a)
+instance (ToRawValue a) => ToRawValue (Vector a)
 
-instance ToRawValue a => ToValue (Vector a) where
+{- ORMOLU_DISABLE -}
+instance (ToRawValue a) => ToValue (Vector a) where
   type NixTypeFor (Vector a) = NixList
   toValue evalState vec =
     coerce <$> do
       let l :: C.CInt
           l = fromIntegral (length vec)
-      v <-
-        [C.block| Value* {
+      b <-
+        [C.block| ListBuilder* {
+#if NIX_IS_AT_LEAST(2, 33, 0)
           EvalState &evalState = *$(EvalState *evalState);
-          Value *v = new (NoGC) Value();
-          evalState.mkList(*v, $(int l));
-          return v;
+          return new (NoGC) ListBuilder(evalState.mem, $(int l));
+#elif NIX_IS_AT_LEAST(2, 32, 0)
+          return new (NoGC) ListBuilder($(int l));
+#else
+          EvalState &evalState = *$(EvalState *evalState);
+          return new (NoGC) ListBuilder(evalState, $(int l));
+#endif
         }|]
       vec & V.imapM_ \i a -> do
         RawValue aRaw <- toRawValue evalState a
         let ix = fromIntegral i
         [C.block| void {
-          Value &v = *$(Value *v);
-          v.listElems()[$(int ix)] = $(Value *aRaw);
+          ListBuilder &b = *$(ListBuilder *b);
+          b[$(int ix)] = $(Value *aRaw);
+        }|]
+      v <-
+        [C.block| Value* {
+          Value * v = new (NoGC) Value();
+          v->mkList(*$(ListBuilder *b));
+          return v;
         }|]
       Value <$> mkRawValue v
+{- ORMOLU_ENABLE -}
 
-instance ToRawValue a => ToRawValue [a]
+instance (ToRawValue a) => ToRawValue [a]
 
-instance ToRawValue a => ToValue [a] where
+instance (ToRawValue a) => ToValue [a] where
   type NixTypeFor [a] = NixList
   toValue es l = toValue es (V.fromList l)
