@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE EmptyDataDecls #-}
 {-# LANGUAGE QuasiQuotes #-}
 {-# LANGUAGE TemplateHaskellQuotes #-}
@@ -9,6 +10,7 @@ module Hercules.CNix.Expr.Context
     Value',
     Attr',
     BindingsBuilder',
+    ListBuilder',
     module Hercules.CNix.Store.Context,
     (=:),
   )
@@ -29,6 +31,10 @@ data Attr'
 
 data BindingsBuilder'
 
+data ListBuilder'
+
+data ListViewType'
+
 context :: C.Context
 context =
   C.cppCtx
@@ -51,4 +57,8 @@ evalContext =
           =: [t|Attr'|]
           <> C.TypeName "BindingsBuilder"
           =: [t|BindingsBuilder'|]
+          <> C.TypeName "ListBuilder"
+          =: [t|ListBuilder'|]
+          <> C.TypeName "ListViewType"
+          =: [t|ListViewType'|]
     }
